#!/usr/bin/env python

# Hut

import os
import pygame
import math
import events
import twister
from pygame.locals import *

from mediator import mediator

NUM_LIFE_STAGES = 6
SIZE = 30
DEFAULT_INIT_HEALTH = 1000
BASE_IMAGE_NAME = "data" + os.path.sep + "Hut30x30_"

# Damage factors
TWIST_STRENGTH_FACTOR = 2
DISTANCE_RADIUS_FACTOR = 2.5    # Number of radiuses a Hut must be inside of to be affected

# Health bar constants
HB_DISPLAY = True
HB_HEIGHT = 4                   # Height of the health bar 
HB_BACK_COLOR = (0, 0, 0)       # Background color of the health bar
HB_SIZE = SIZE - 8              # Size of the Health bar
HB_FRONT_COLOR = (255, 0, 0)    # Color of the health in the health bar
HB_HUT_DISTANCE = -2            # Distance between the health bar and the Hut

class Hut(pygame.sprite.Sprite):

    img_array = []

    def __init__(self, initPos, initHealth=DEFAULT_INIT_HEALTH, mediator=mediator):
        """
        Create a new Hut

        `initPos` 2-tuple (X, Y) defining the Hut location
        `initHealth` Initial (maximum) health of object (optional) 
        """
        
        super(Hut, self).__init__()

        self.posX = initPos[0]
        self.posY = initPos[1]
        self.hb_display = HB_DISPLAY
        self._image = None
        self.mediator = mediator
        
        # Set internal variables
        self.__size = SIZE
        self.__initHealth = initHealth
        self.__health = self.__initHealth
        if (self.__initHealth < NUM_LIFE_STAGES):
            self.__initHealth = NUM_LIFE_STAGES    
        self.__stage_dec = self.__initHealth/(NUM_LIFE_STAGES - 1)

        if (not Hut.img_array):
            for i in range(0, NUM_LIFE_STAGES):
                Hut.img_array.append(pygame.image.load(BASE_IMAGE_NAME + `i` + ".png"))
            Hut.img_dead = pygame.image.load(BASE_IMAGE_NAME + "dead.png")

        self.rect = pygame.rect.Rect(0, 0, self.__size, self.__size + 10) 
        self.rect.center = (self.posX, self.posY)

        self.mediator.addListener(events.TWISTER_MOVED, self.handleTwisterMoved)
        
    def health():
        doc = "Health level of the Hut"
        def fget(self):
            return self.__health
        def fset(self, newHealth):
            self.__health = newHealth
            if (self.__health < 0):
                self.__health = 0
            elif (self.__health > self.__initHealth):
                self.__health = self.__initHealth
            self._image = None
            # Handle hut death
            if (self.__health == 0):
                self.mediator.removeListener(
                    events.TWISTER_MOVED, self.handleTwisterMoved)
                self.mediator.publishEvent('hut_died', hut=self)
        return locals()
    health = property(**health())

    def getHealthBarImages(self):
        hb_back_img = pygame.Surface([HB_SIZE, HB_HEIGHT])
        hb_back_img.fill(HB_BACK_COLOR)

        topleft = [(self.__size - HB_SIZE)/2, self.__size + HB_HUT_DISTANCE]

        hb_back_rect = hb_back_img.get_rect()
        hb_back_rect.topleft = topleft

        hb_front_img = pygame.Surface([(HB_SIZE * self.health)/self.__initHealth, HB_HEIGHT])
        hb_front_img.fill(HB_FRONT_COLOR)

        hb_front_rect = hb_front_img.get_rect()
        hb_front_rect.topleft = topleft

        return [(hb_back_img, hb_back_rect), (hb_front_img, hb_front_rect)]

    def handleTwisterMoved(self, event):
        twister = event.twister

        # Calculate distance
        xDelta = twister.position.x - self.posX
        yDelta = twister.position.y - self.posY
        distance = math.sqrt(xDelta*xDelta + yDelta*yDelta)
        
        # (twist*TWIST_STRENGTH_FACTOR)^radius/distance
        if (distance < twister.radius*DISTANCE_RADIUS_FACTOR):
            health_decrement = math.pow(math.fabs(twister.twist)*TWIST_STRENGTH_FACTOR, twister.radius/distance)
            self.health -= int(health_decrement)
        
    def image():
        doc = "Sprite image"
        def fget(self):
            if (self._image):
                return self._image
            
            if (0 == self.health):
                main_img = Hut.img_dead
                next_img = None
            else:
                # Calculate the image index (image at index 0 should be the original hut and 
                # hut images should crumble as the index rises)
                stage_increments = float(self.__initHealth)/float(NUM_LIFE_STAGES)
                current_stage = int(self.health/stage_increments)
                max_index = NUM_LIFE_STAGES - 1
                if (current_stage > max_index):
                    current_stage = max_index 
                image_index = int(math.fabs(current_stage - max_index))

                # Set the main Hut image
                main_img = Hut.img_array[image_index] 

                # Set the next image in the animation
                if (image_index < (max_index)):
                    next_img = Hut.img_array[image_index + 1]
                else:
                    next_img = Hut.img_dead

                if (self.health < self.__initHealth):
                    # Calculate alpha on the main image
                    # THIS IS NOT THREAD SAFE!!!
                    a = self.health/stage_increments
                    a = a - int(a)
                    alpha = a*255
                    main_img.set_alpha(alpha)
            
            # Get surface
            _surf = pygame.surface.Surface(
                (self.__size, self.__size + 10)).convert_alpha()
            _surf.fill((0,0,0,0))
            
            # Render the Hut
            if (next_img is not None):
                _surf.blit(next_img, (0,0))
            _surf.blit(main_img, (0,0))

            # Reset alpha on the main image
            main_img.set_alpha(255)

            if (self.hb_display):
                # Render the Health bar
                health_bar_images = self.getHealthBarImages()
                for img_tuple in health_bar_images:
                    _surf.blit(img_tuple[0], img_tuple[1])
            self._image = _surf
            return _surf
        return locals()
    image = property(**image())

# Some test code to animate
if __name__=='__main__':
    # hut_array is [Hut object, whether it is building or crumbling]
    hut_array = []
    hut_array.append(Hut((100, 100)))
    hut_array.append(Hut((136, 92)))
    hut_array.append(Hut((230, 15)))
    hut_array.append(Hut((18, 93)))
    hut_array.append(Hut((211, 156)))
    hut_array.append(Hut((98, 15)))
    hut_array.append(Hut((53, 211)))
    hut_array.append(Hut((200, 89)))
    hut_array.append(Hut((67, 178)))
    hut_array.append(Hut((83, 192)))
    hut_array.append(Hut((26, 124)))
    
    bg = pygame.image.load(os.path.join("data","bgtest.jpg"))
    screen = pygame.display.set_mode([800, 600])
    thereIsStillAHutAlive = True
    t = twister.Twister((0, 0), 50, moveVector=(5,5), twist=30)
    while (pygame.event.poll().type != KEYDOWN):
        thereIsStillAHutAlive = False
        pygame.time.delay(10)
        mediator.publishEvent('tick')
        screen.blit(bg, (0,0))
        for hut in hut_array:
            screen.blit(hut.image, hut.rect)
        screen.blit(t.image, t.rect)
        pygame.display.update()
